/* Copyright (C) 2009 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "data.h"

#include<iostream>
#include<iomanip>
#include<vector>
#include<string>
#include<cmath>
#include<gsl/gsl_randist.h>
#include<gsl/gsl_cdf.h>

using namespace std;

Data::Data(string datafilename)
{
  _obs_status = NULL;
  _datafilename=datafilename;
  vector<string> pos;
  vector<float> x;
  vector<float> rx;
  vector<float> seqs;
  
  string posi;
  float xi;
  float rxi;
  float seqsi;
  

  ifstream input (datafilename.c_str(), ios::in);
  while(input.get()!='\n') {}
  while (!input.eof()) {
    input >> posi >> xi >> rxi >> seqsi;
    //cout << posi << " " << xi << " " << rxi << " " << seqsi << endl;
    if (!input.eof()) {
      pos.push_back(posi);
      x.push_back(xi);
      rx.push_back(rxi);
      seqs.push_back(seqsi);
    }
  }
  input.close();
  _lseq = x.size();
  cout << "lseq " << _lseq << endl; 
  cout << "pos " << pos[0] << " x " << x[0] << " rx " << rx[0] << " seqs " << seqs[0] << endl;
  cout << "pos " << pos[_lseq-1] << " x " << x[_lseq-1] << " rx " << rx[_lseq-1]<< " seqs " << seqs[_lseq-1] << endl;

  _pos = new string [_lseq];
  _x = new float [_lseq];
  _rx = new float [_lseq];
  _seqs = new float [_lseq];
  for (int i=0; i<_lseq; i++) {
    _pos[i]=pos[i];
    _x[i]=x[i];
    _rx[i]=rx[i];
    _seqs[i]=seqs[i];
  }
  
  _obs_status = new int [_lseq];

  for (int i=0; i<_lseq; i++) {
    _obs_status[i]=1;
    if (_x[i] < 0) {
      _obs_status[i]=0;
    }
  }
}

void Data::allocalgovar(int nS, int nR, float xmin, float xmax, float absrxmax, float seqsmax)
{
  _nR = nR;
  _nS = nS;
  _xmin = xmin;
  _xmax = xmax;
  
  for (int i=0; i<_lseq; i++) {
    if (_x[i]<_xmin) {
      _obs_status[i]=2;
    }
    if (_x[i]>_xmax) {
      _obs_status[i]=3;
    }
    if (_rx[i]<-absrxmax) {
      _obs_status[i]=0;
    }
    if (_rx[i]>absrxmax) {
      _obs_status[i]=0;
    }
    if (_seqs[i]>seqsmax) {
      _obs_status[i]=0;
    }
  } 
  
  _CIlevel = 0.95;
  _mostlikelyS = new int [_lseq];
  _mostlikelyR = new int [_lseq];
  _expectS = new float [_lseq];
  _postpoutlier = new float [_lseq];
  _minCI = new int [_lseq];
  _maxCI = new int [_lseq];

  _viterbiR = new int [_lseq];
  _viterbiM = new int [_lseq];
  _viterbiS = new int [_lseq];

  _posteriorM1u = new float [_lseq];
  _posteriorM1d = new float [_lseq];
  _posteriorM2 = new float [_lseq];
  _posteriorM3 = new float [_lseq];

  _posteriorR = new float* [_lseq];
  for (int i=0; i<_lseq; i++) {
    _posteriorR[i] = new float [_nR];
  }
}

vector<float> Data::range()
{
  vector<float> retval;
  retval.push_back(_x[0]);
  retval.push_back(_x[0]);
  for (int i=0; i<_lseq; i++) {
    if (_x[i]<retval[0]) {
      retval[0]=_x[i];
    }
    if (_x[i]>retval[1]) {
      retval[1]=_x[i];
    }
  }
  return(retval);
}

float Data::absrxmax()
{
  float rmax = fabs(_rx[0]);
  for (int i=0; i<_lseq; i++) {
    if (fabs(_rx[i]) > rmax) {
      rmax = fabs(_rx[i]);
    }
  }
  return(rmax);
}

int Data::nxinfmu(float mu)
{
  int n=0;
  for (int i=0; i<_lseq; i++) {
    if (_obs_status[i]!=0 && _x[i]<mu) {
      n++;
    }
  } 
  return(n);
}

void Data::VIT(Params& par, FbVar& fbv, int estep_l, int estep_ov)
{
  int ltot=_lseq;
  int irecmax=-1;
  for (int imin=0; irecmax<ltot-1; imin+=estep_l-estep_ov) {
    int imax=imin+estep_l-1;
    int irecmin=irecmax+1;
    irecmax=imax-estep_ov/2;
    if (imax>ltot-1) {
      imax=ltot-1;
    }
    if (irecmax>ltot-1) {
      irecmax=ltot-1;
    }
    this->viterbi(par,fbv,imin,imax,irecmin,irecmax);
  }
}

void Data::viterbi(Params& par, FbVar& fbv, int imin, int imax, int irecmin, int irecmax)
{
  float* VIval = new float [_nS];
  float* VDval = new float [_nS];
  int* VIbackS = new int [_nS];
  int* VDbackS = new int [_nS];

  int lalgo=imax-imin+1;
  for (int t=imin; t<=imax; t++) {
    for (int u=0; u<_nS; u++) {
      if (_obs_status[t]==0) {
	fbv._pobs_gaussian[t-imin][u]= 1;
      } else if (_obs_status[t]==1) {
	fbv._pobs_gaussian[t-imin][u]=gsl_ran_gaussian_pdf(_x[t]-par._mu[u]-par._rho_v[u]*_rx[t],par._sigma_v[u]);
      } else if (_obs_status[t]==2) {
	fbv._pobs_gaussian[t-imin][u]=gsl_cdf_gaussian_P(_xmin-par._mu[u]-par._rho_v[u]*_rx[t],par._sigma_v[u]);
      } else if (_obs_status[t]==3) {
	fbv._pobs_gaussian[t-imin][u]=gsl_cdf_gaussian_Q(_xmax-par._mu[u]-par._rho_v[u]*_rx[t],par._sigma_v[u]);
      }      
    }    
  }

  // initialisation
  if (imin==irecmin) {
    for (int p=0; p<_nR; p++) {
      for (int u=0; u<_nS; u++) {
	fbv._vitVval[0][p][u] = 0;
	fbv._vitWval[0][p][u] = log(par._r0[p]*par._s0[u]);
	fbv._vitVbackM[0][p][u] = -1;
	fbv._vitVbackS[0][p][u] = -1;
	fbv._vitWbackR[0][p][u] = -1;
      }
      double poutl = par._poutlierlogit[(int)(fabs(_rx[imin])*100)];
      for (int u=0; u<_nS; u++) {
	//float x=log(fbv._pobs_gaussian[0][u]);
	float x=log(poutl*par._pobs_outlier[_obs_status[0+imin]] +(1-poutl)*fbv._pobs_gaussian[0][u]);
	if (isfinite(x) && x>-1000) {
	  fbv._vitWval[0][p][u]+=x;
	} else {
	  fbv._vitWval[0][p][u]+=-1000;
	}
      }
    }
  } else {
    for (int p=0; p<_nR; p++) {
      for (int u=0; u<_nS; u++) {
	fbv._vitVval[0][p][u] = 0;
	fbv._vitWval[0][p][u] = -1000;
	fbv._vitVbackM[0][p][u] = -1;
	fbv._vitVbackS[0][p][u] = -1;
	fbv._vitWbackR[0][p][u] = -1;
      }
    }
    fbv._vitWval[0][_viterbiR[imin]][_viterbiS[imin]] = 0;
  }

    if (0 && imin < 100) {
      int t=0;
      for (int p=0; p<_nR; p++) {
	cout << "fbv._vitVval[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitVval[t][p][v];
	}
	cout << endl;
	cout << "fbv._vitWval[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitWval[t][p][v];
	}
	cout << endl;
	cout << "fbv._vitWbackR[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitWbackR[t][p][v];
	}
	cout << endl;
	cout << "fbv._vitVbackM[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitVbackM[t][p][v];
	}
	cout << endl;
	cout << "fbv._vitVbackS[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitVbackS[t][p][v];
	}
	cout << endl;
      }
    }

 
  //Recurrence Viterbi Forward

  for (int t=1; t<lalgo; t++) {

    for (int p=0; p<_nR; p++) {

      VIval[0] = -1000;
      VIbackS[0] = -1;
      for (int v=1; v<_nS; v++) {
	VIval[v] = log(par._lambda_i[p]) + VIval[v-1];
	VIbackS[v] = VIbackS[v-1];
	if (fbv._vitWval[t-1][p][v-1] > VIval[v] || VIbackS[v]<0) {
	  VIval[v] = fbv._vitWval[t-1][p][v-1];
	  VIbackS[v] = v-1;
	}
      }


      VDval[_nS-1] = -1000;
      VDbackS[_nS-1] = -1;
      for (int v=_nS-2; v>=0; v--) {
	VDval[v] = log(par._lambda_d[p]) + VDval[v+1];
	VDbackS[v] = VDbackS[v+1];
	if (fbv._vitWval[t-1][p][v+1] > VDval[v] || VDbackS[v]<0) {
	  VDval[v] = fbv._vitWval[t-1][p][v+1];
	  VDbackS[v] = v+1;
	}
      }

      int umax = 0;
      for (int u=1; u<_nS; u++) {
	if (fbv._vitWval[t-1][p][u] > fbv._vitWval[t-1][p][umax]) {
	  umax=u;
	}
      }     

      for (int v=0; v<_nS; v++) {
	float a=log(1-par._lambda_i[p]);
	float b=log(1-par._lambda_d[p]);
	if (v==_nS-1) {
	  a = 0;
	}
	if (v==0) {
	  b = 0;
	}
	fbv._vitVval[t][p][v] = log(par._alpha[p][0]) + fbv._vitWval[t-1][p][v];
	fbv._vitVbackS[t][p][v] = v;
	fbv._vitVbackM[t][p][v] = 0;
	
	float x1 = log(par._alpha[p][1]*par._fstat[v]) + fbv._vitWval[t-1][p][umax];
	//float x1 = log(par._alpha[p][1]/(float)(_nS)) + fbv._vitWval[t-1][p][umax];
	if (x1 > fbv._vitVval[t][p][v]) {
	  fbv._vitVval[t][p][v] = x1;
	  fbv._vitVbackS[t][p][v] = umax;
	  fbv._vitVbackM[t][p][v] = 1;
	}

	float x2 = log(par._alpha[p][2]) + a + VIval[v];
	if (x2 > fbv._vitVval[t][p][v] && VIbackS[v]>=0) {
	  fbv._vitVval[t][p][v] = x2;
	  fbv._vitVbackS[t][p][v] = VIbackS[v];
	  fbv._vitVbackM[t][p][v] = 2;
	}

	float x3 = log(par._alpha[p][3]) + b + VDval[v];
	if (x3 > fbv._vitVval[t][p][v] && VDbackS[v]>=0) {
	  fbv._vitVval[t][p][v] = x3;
	  fbv._vitVbackS[t][p][v] = VDbackS[v];
	  fbv._vitVbackM[t][p][v] = 3;
	}

	if (v==_nS-1) {
	  float x2bis = log(par._alpha[p][2]) + fbv._vitWval[t-1][p][v];
	  if (x2bis > fbv._vitVval[t][p][v]) {
	    fbv._vitVval[t][p][v] = x2bis;
	    fbv._vitVbackS[t][p][v] = v;
	    fbv._vitVbackM[t][p][v] = 2;
	  }	  
	}
	if (v==0) {
	  float x3bis = log(par._alpha[p][3]) + fbv._vitWval[t-1][p][v];
	  if (x3bis > fbv._vitVval[t][p][v]) {
	    fbv._vitVval[t][p][v] = x3bis;
	    fbv._vitVbackS[t][p][v] = v;
	    fbv._vitVbackM[t][p][v] = 3;
	  }
	}

	//float x=log(fbv._pobs_gaussian[t][v]);
	double poutl = par._poutlierlogit[(int)(fabs(_rx[t+imin])*100)];
	float x=log(poutl*par._pobs_outlier[_obs_status[t+imin]] +(1-poutl)*fbv._pobs_gaussian[t][v]);
	if (isfinite(x) && x>-1000) {
	  fbv._vitVval[t][p][v]+=x;
	} else {
	  fbv._vitVval[t][p][v]+=-1000;
	}
      }
    }

    for (int p=0; p<_nR; p++) {
      for (int v=0; v<_nS; v++) {
	for (int pbef=0; pbef<_nR; pbef++) {
	  float x = log(par._b[pbef][p]) + fbv._vitVval[t][pbef][v];
	  if (pbef==0 || x > fbv._vitWval[t][p][v]) {
	    fbv._vitWval[t][p][v] = x;
	    fbv._vitWbackR[t][p][v] = pbef;
	  }
	}
      } 
    }
    if (0 && t+imin < 100) {
//       cout << "VDval";
//       for (int u=0; u<_nS; u++) {
// 	cout << " " << VDval[u];
//       }
//       cout << endl;
//       cout << "VDbackS";
//       for (int u=0; u<_nS; u++) {
// 	cout << " " << VDbackS[u];
//       }
      cout << endl;
      for (int p=0; p<_nR; p++) {
	cout << "fbv._vitVval[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitVval[t][p][v];
	}
	cout << endl;
	cout << "fbv._vitWval[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitWval[t][p][v];
	}
	cout << endl;
	cout << "fbv._vitWbackR[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitWbackR[t][p][v];
	}
	cout << endl;
	cout << "fbv._vitVbackM[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitVbackM[t][p][v];
	}
	cout << endl;
	cout << "fbv._vitVbackS[" << t+imin << "][0]";
	for (int v=0; v<_nS; v++) {
	  cout << " " << fbv._vitVbackS[t][p][v];
	}
	cout << endl;
      }
    }
    
  }

  //Backtracking
  int Mmax=0;
  int Rmax=0;
  int Smax=0;
  for (int p=0; p<_nR; p++) {
    for (int v=0; v<_nS; v++) {
      if (fbv._vitWval[lalgo-1][p][v] > fbv._vitWval[lalgo-1][Rmax][Smax]) {
	Rmax = p;
	Smax = v;
      }
    }
  }
  
  for (int t=lalgo-1; t>0; t--) {
    if (t+imin>=irecmin && t+imin<=irecmax) {
      _viterbiR[t+imin] = Rmax;
      _viterbiS[t+imin] = Smax;
      _viterbiM[t+imin] = Mmax;
    }

    Rmax = fbv._vitWbackR[t][Rmax][Smax];
    int Smaxold = Smax;
    Smax = fbv._vitVbackS[t][Rmax][Smaxold];
    Mmax = fbv._vitVbackM[t][Rmax][Smaxold];    
  }
  
  if (imin>=irecmin && imin<=irecmax) {
    _viterbiR[imin] = Rmax;
    _viterbiS[imin] = Smax;
    _viterbiM[imin] = Mmax;
  }


  delete VIval;
  delete VDval;
  delete VIbackS;
  delete VDbackS;
}

double Data::FB(Params& par, FbVar& fbv, int estep_l, int estep_ov)
{
  double loglik = 0;
  int ltot=_lseq;
  int irecmax=-1;
  for (int imin=0; irecmax<ltot-1; imin+=estep_l-estep_ov) {
    int imax=imin+estep_l-1;
    int irecmin=irecmax+1;
    irecmax=imax-estep_ov/2;
    if (imax>ltot-1) {
      imax=ltot-1;
    }
    if (irecmax>ltot-1) {
      irecmax=ltot-1;
    }
    //cout << "imin " << imin << " imax " << imax << " irecmin " << irecmin << " irecmax " << irecmax << endl;
    loglik += this->forward(par,fbv,imin,imax,irecmin,irecmax);
    this->backward(par,fbv,imin,imax,irecmin,irecmax);
  }
  return(loglik);
}


double Data::forward(Params& par, FbVar& fbv, int imin, int imax, int irecmin, int irecmax)
{
  double loglik = 0;
  double tsumloglik;
  double tsum, tsum1, tsum2, tsum3;
  double* IF = new double [_nS];
  double* DF = new double [_nS];

  int lalgo=imax-imin+1;
  for (int t=imin; t<=imax; t++) {
    for (int u=0; u<_nS; u++) {
      if (_obs_status[t]==0) {
	fbv._pobs_gaussian[t-imin][u]= 1;
      } else if (_obs_status[t]==1) {
	fbv._pobs_gaussian[t-imin][u]=gsl_ran_gaussian_pdf(_x[t]-par._mu[u]-par._rho_v[u]*_rx[t],par._sigma_v[u]);
      } else if (_obs_status[t]==2) {
	fbv._pobs_gaussian[t-imin][u]=gsl_cdf_gaussian_P(_xmin-par._mu[u]-par._rho_v[u]*_rx[t],par._sigma_v[u]);
      } else if (_obs_status[t]==3) {
	fbv._pobs_gaussian[t-imin][u]=gsl_cdf_gaussian_Q(_xmax-par._mu[u]-par._rho_v[u]*_rx[t],par._sigma_v[u]);
      }      
    }    
  }

  //cout << "Data::forward1 _minCI " << _minCI << endl;
  // initialisation
  tsumloglik=0;
  double poutl = par._poutlierlogit[(int)(fabs(_rx[0+imin])*100)];
  for (int p=0; p<_nR; p++) {
    tsum = 0;
    for (int u=0; u<_nS; u++) {
      fbv._pforw1[0][p][u]=par._r0[p]*par._fstat[u]; //*par._s0[u];
      fbv._pforw2[0][p][u]= fbv._pforw1[0][p][u]*(poutl*par._pobs_outlier[_obs_status[0+imin]] +(1-poutl)*fbv._pobs_gaussian[0][u]);

      tsum += fbv._pforw2[0][p][u];
    }
    tsumloglik += tsum;
  }

  for (int p=0; p<_nR; p++) {
    for (int u=0; u<_nS; u++) {
      fbv._pforw2[0][p][u]=fbv._pforw2[0][p][u]/tsumloglik;
    }
  }
  
  if (imin>=irecmin) {
    loglik += log(tsumloglik);
  }
  for (int p=0; p<_nR; p++) {
    for (int u=0; u<_nS; u++) {
      tsum1=0;
      tsum2=0;
      for (int pbef=0; pbef<_nR; pbef++) {
	tsum1 += fbv._pforw2[0][pbef][u]*par._b[pbef][p];
	tsum2 += fbv._pforw1[0][pbef][u]*par._b[pbef][p];
      }
      fbv._pforw3[0][p][u] = tsum1;
      fbv._pforw4[0][p][u] = tsum2;
    }
  }

    if (0) {
      int t=0;
      for (int p=0; p<_nR; p++) {
	cout << "p " << p << endl;
	tsum1=0;
	tsum2=0;
	tsum3=0;
	cout << "obs_status " << _obs_status[t+imin] << " _x "<< _x[t+imin] << " _x-_rx "<< _x[t+imin]-_rx[t+imin] << endl;
	for (int u=0; u<_nS; u++) {
	  tsum1+=fbv._pforw1[t][p][u];
	  tsum2+=fbv._pforw2[t][p][u];
	  tsum3+=fbv._pforw3[t][p][u];
	}
	cout <<" _pforw1["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pforw1[t][p][u];
	}
	cout << endl;
	cout <<" _pforw2["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pforw2[t][p][u];
	}
	cout << endl;
	cout <<" _pforw3["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pforw3[t][p][u];
	}
	cout << endl;
	cout << "tsum1 " << tsum1-0.5 << " tsum2 " << tsum2-0.5 << " tsum3 " << tsum3-0.5 << endl;
      }
    }
 
  //cout << "Data::forward2 _minCI " << _minCI << endl;

  for (int t=1; t<lalgo; t++) {
    tsumloglik=0;
    for (int p=0; p<_nR; p++) {

      IF[0] = 0;
      for (int v=1; v<_nS; v++) {
	IF[v]=par._lambda_i[p]*IF[v-1] + fbv._pforw3[t-1][p][v-1];
      }
      DF[_nS-1] = 0;
      for (int v=_nS-2; v>=0; v--) {
	DF[v]=par._lambda_d[p]*DF[v+1] + fbv._pforw3[t-1][p][v+1];
      }

      double pp=0;
      for (int v=0; v<_nS; v++) {
	pp += fbv._pforw3[t-1][p][v];
      }

      tsum = 0;
      for (int v=0; v<_nS; v++) {
	double a=1-par._lambda_i[p];
	double b=1-par._lambda_d[p];
	if (v==_nS-1) {
	  a = 1;
	}
	if (v==0) {
	  b = 1;
	}
	fbv._pforw1[t][p][v]=par._alpha[p][0]*fbv._pforw3[t-1][p][v] 
	  //+ (par._alpha[p][1]/(double)(_nS))*pp 
	  + par._alpha[p][1]*par._fstat[v]*pp 
	  + par._alpha[p][2]*a*IF[v] + par._alpha[p][3]*b*DF[v];
	if (v==_nS-1) {
	  fbv._pforw1[t][p][v]+= par._alpha[p][2]*fbv._pforw3[t-1][p][v];
	}
	if (v==0) {
	  fbv._pforw1[t][p][v]+= par._alpha[p][3]*fbv._pforw3[t-1][p][v];
	}
// 	if (v==0 && t+imin <= 10) {
// 	  cout << "p=" << p << " " << par._alpha[p][0]*fbv._pforw3[t-1][p][v] << " + " << (par._alpha[p][1]/(double)(_nS))*pp << " + " << par._alpha[p][2]*a*IF[v] + par._alpha[p][3]*b*DF[v] << endl;
// 	}
	poutl = par._poutlierlogit[(int)(fabs(_rx[t+imin])*100)];
	fbv._pforw2[t][p][v]= fbv._pforw1[t][p][v]*(poutl*par._pobs_outlier[_obs_status[t+imin]] +(1-poutl)*fbv._pobs_gaussian[t][v]);
	tsum += fbv._pforw2[t][p][v];
      }
      tsumloglik += tsum;
      
    }

    for (int p=0; p<_nR; p++) {
      for (int v=0; v<_nS; v++) {
	fbv._pforw2[t][p][v]=fbv._pforw2[t][p][v]/tsumloglik;
      } 
    }
    
    if (t+imin>=irecmin && t+imin<=irecmax) {
      loglik += log(tsumloglik);
    }
    
    for (int p=0; p<_nR; p++) {
      for (int u=0; u<_nS; u++) {
	tsum1=0;
	tsum2=0;
	for (int pbef=0; pbef<_nR; pbef++) {
	  tsum1 += fbv._pforw2[t][pbef][u]*par._b[pbef][p];
	  tsum2 += fbv._pforw1[t][pbef][u]*par._b[pbef][p];
	}
	fbv._pforw3[t][p][u] = tsum1;
	fbv._pforw4[t][p][u] = tsum2;
      }
    }
    
    if (0 && t+imin<10) {
      for (int p=0; p<_nR; p++) {
	cout << "p " << p << endl;
	tsum1=0;
	tsum2=0;
	tsum3=0;
	cout << "obs_status " << _obs_status[t+imin] << " _x "<< _x[t+imin] << " _x-_rx "<< _x[t+imin]-_rx[t+imin] << endl;
	for (int u=0; u<_nS; u++) {
	  tsum1+=fbv._pforw1[t][p][u];
	  tsum2+=fbv._pforw2[t][p][u];
	  tsum3+=fbv._pforw3[t][p][u];
	}
	cout <<" _pforw1["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pforw1[t][p][u];
	}
	cout << endl;
	cout <<" _pforw2["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pforw2[t][p][u];
	}
	cout << endl;
	cout <<" _pforw3["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pforw3[t][p][u];
	}
	cout << endl;
	cout <<" _pforw4["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pforw4[t][p][u];
	}
	cout << endl;
	cout << "tsum1 " << tsum1-0.5 << " tsum2 " << tsum2-0.5 << " tsum3 " << tsum3-0.5 << endl;
      }
    }
    
  }
  

  delete IF;
  delete DF;
  
  return(loglik);
}

void Data::count_sum_sd(Params& par,int t, float** pvect, float* pvectobs)
{
  //cout << t << " Data::count_sum_sd" << endl;
  double eps = 1e-30;
  double tsum=0;
  double tsumout=0;
  double poutl = par._poutlierlogit[(int)(fabs(_rx[t])*100)];
  double pobsout = poutl*par._pobs_outlier[_obs_status[t]];

  for (int u=0; u<_nS; u++) {

    float prob = 0;
    for (int p=0; p<_nR; p++) {
      prob += pvect[p][u];
    }
    //cout << "prob u=" << u << " " << prob << endl;
    if (prob>eps) {
      double pobstot = pobsout + (1-poutl)*pvectobs[u];
      double pout = pobsout/pobstot;
      tsumout += pout*prob;
    
      double tp=(1-pout)*prob;
      if (tp>eps) {
	par._count_sum_rhosd_0[u] += tp;
	par._count_sum_rhosd_1[u] += tp*pow(_rx[t],2);
	
	if (_obs_status[t]==0) {
	  
	  tsum += tp*par._sigma2_v[u];
	  
	  par._count_sum_rhosd_2[u] += tp*_rx[t]*(_rx[t]*par._rho_v[u]);
	  par._count_sum_rhosd_3[u] += tp*(par._sigma2_v[u]+pow(_rx[t],2)*pow(par._rho_v[u],2));
	  
	} else if (_obs_status[t]==1) {
	  
	  tsum += tp*pow((_x[t]-par._mu[u]-par._rho_v[u]*_rx[t]),2);
	  
	  par._count_sum_rhosd_2[u] += tp*_rx[t]*(_x[t]-par._mu[u]);
	  par._count_sum_rhosd_3[u] += tp*pow(_x[t]-par._mu[u],2);
	  
	}  else if (_obs_status[t]==2) {
	  
	  double x = (_xmin-par._mu[u]-par._rho_v[u]*_rx[t])/par._sigma_v[u];
	  double fx=gsl_ran_gaussian_pdf(x,1)/gsl_cdf_gaussian_P(x,1);
	  
	  if (!isnan(fx)) {
	    tsum += tsum*par._sigma2_v[u]*(1-x*fx);
	    
	    par._count_sum_rhosd_2[u] += tp*_rx[t]*(par._rho_v[u]*_rx[t] - par._sigma_v[u]*fx);
	    par._count_sum_rhosd_3[u] += tp*(pow(_rx[t]*par._rho_v[u],2) -2*_rx[t]*par._rho_v[u]*par._sigma_v[u]*fx + par._sigma2_v[u]*(1-x*fx));
	  }
	} else if (_obs_status[t]==3) {
	  
	  double x = (_xmax-par._mu[u]-par._rho_v[u]*_rx[t])/par._sigma_v[u];
	  double fx=gsl_ran_gaussian_pdf(x,1)/gsl_cdf_gaussian_Q(x,1);
	  if (!isnan(fx)) {
	    tsum += tp*par._sigma2_v[u]*(1+x*fx);
	    
	    par._count_sum_rhosd_2[u] += tp*_rx[t]*(par._rho_v[u]*_rx[t] + par._sigma_v[u]*fx);
	    par._count_sum_rhosd_3[u] += tp*(pow(_rx[t]*par._rho_v[u],2) +2*_rx[t]*par._rho_v[u]*par._sigma_v[u]*fx + par._sigma2_v[u]*(1+x*fx));
	  }
	}  
      }
    }
  }
  par._count_sum_sd += tsum;
  par._count_outlier += tsumout;
  _postpoutlier[t] = tsumout;
  if (par._type_outlier == 2) {
    int j = (int)(100*_rx[t]);
    if (j<0) {j=-j;}
    par._count_logitoutlier[j][0] += tsumout;
    par._count_logitoutlier[j][1] += (1-tsumout);
  }
}

void Data::count_t(Params& par, int t)
{
  par._count_t++;
}

void Data::CIpos(int t, float** pvect, Params& par)
{
  float sump;

  int umax = 0;
  _minCI[t]=-1;
  _maxCI[t]=-1;
  _expectS[t]=0;
  sump=0;
  float probmax = -1;
  for (int u=0; u<_nS; u++) {
    float prob = 0;
    for (int p=0; p<_nR; p++) {
      prob += pvect[p][u];
    }
    _expectS[t] += prob*par._mu[u];
    sump += prob;
    if (_minCI[t]==-1 && sump >= (1-_CIlevel)/2) {
      _minCI[t] = u;
    }
    if (_maxCI[t]==-1 && sump >= (1+_CIlevel)/2) {
      _maxCI[t] = u;
    }
    if (prob>probmax) {
      probmax=prob;
      umax=u;
    }
  }
  if (_maxCI[t]==-1) {
    _maxCI[t] = _nS-1;
  }
  _mostlikelyS[t]=umax;

  int qmax=0;
  float pqmax=0;
  for (int q=0; q<_nR; q++) {
    float prob=0;
    for (int u=0; u<_nS; u++) {
      prob += pvect[q][u];
    }
    _posteriorR[t][q]=prob;
    if (prob>pqmax) {
      pqmax=prob;
      qmax=q;
    }
  }
  _mostlikelyR[t] = qmax;
}

void Data::printPath(Params& par, ofstream& out)
{
  out << "Xstatus" << "\t" << "t" << "\t" << "SeqS" << "\t" << "X" << "\t" << "XrX" << "\t" << "XrrX" << "\t" << "Poutl" << "\t" << "ES" << "\t" << "MLS" << "\t" << "CIminS" << "\t" << "CImaxS";
  for (int p=0; p<_nR; p++) {
    out << "\t" << "PR" << p;
  }
  out << "\t" << "PM1" << "\t" << "PM2" << "\t" << "PM3" << "\t" << "VitR" << "\t" << "VitM" << "\t" << "VitS" << "\t" << "PM1up" << "\t" << "PM1down" << endl;
  for (int t=0; t<_lseq; t++) {
    out << _obs_status[t] << "\t" << t << "\t" << _seqs[t] << "\t" << _x[t] << "\t" << _x[t]-_rx[t] << "\t" << _x[t]-par._rho_v[_mostlikelyS[t]]*_rx[t] << "\t" << _postpoutlier[t] << "\t" << _expectS[t] << "\t" << par._mu[_mostlikelyS[t]] << "\t" << par._mu[_minCI[t]] << "\t" << par._mu[_maxCI[t]];
    for (int p=0; p<_nR; p++) {
      out << "\t" << _posteriorR[t][p];
    }
    out << "\t" << 0.0001*(int)((_posteriorM1u[t]+_posteriorM1d[t])*10000) << "\t" << 0.0001*(int)(_posteriorM2[t]*10000) << "\t" << 0.0001*(int)(_posteriorM3[t]*10000); 
    out << "\t" << _viterbiR[t] << "\t" << _viterbiM[t] << "\t" << par._mu[_viterbiS[t]];
    out << "\t" << 0.0001*(int)(_posteriorM1u[t]*10000) << "\t" << 0.0001*(int)(_posteriorM1d[t]*10000) << endl;
  }
}

void Data::backward(Params& par, FbVar& fbv, int imin, int imax, int irecmin, int irecmax)
{
  int t;

  double* B = new double [_nS];
  double* Bfcum = new double [_nS];
  double* IB = new double [_nS];
  double* DB = new double [_nS];
  double* tmp_count_alpha = new double [5]; 
  double* IC = new double [_nS];
  double* ID = new double [_nS];
  double* DC = new double [_nS];
  double* DD = new double [_nS];

  double eps = 1e-30;

  int lalgo = imax-imin+1;
  t=lalgo-1;
  for (int p=0; p<_nR; p++) {
    for (int u=0; u<_nS; u++) {
      fbv._pback1[t][p][u]=fbv._pforw3[t][p][u];
    }    
  }
  for (int p=0; p<_nR; p++) {
    //double tsum=0;
    for (int v=0; v<_nS; v++) {
      fbv._pback2[t][p][v]=0;
      for (int q=0; q<_nR; q++) {
	if (fbv._pforw3[t][p][v] > eps) {
	  fbv._pback2[t][p][v]+=(fbv._pforw2[t][p][v]*par._b[p][q]*fbv._pback1[t][q][v])/fbv._pforw3[t][p][v];
	} 
      }
      //tsum+=fbv._pback2[t][p][v];
    } 
    //cout << "tsum fbv._pback2 " << tsum << endl; 
  }

  if (t+imin>=irecmin && t+imin<=irecmax) {
    this->CIpos(t+imin, fbv._pback1[t], par);
    this->count_t(par,t+imin);
    this->count_sum_sd(par,t+imin, fbv._pback1[t],fbv._pobs_gaussian[t]);
  }

  if (0) {
    for (int p=0; p<_nR; p++) {
      double tsum1=0;
      double tsum2=0;
      cout << "_x "<< _x[t] <<" fbv._pback1["<< t<<"][" <<p<<"] "; 
      for (int u=0; u<_nS; u++) {
	cout << " " << fbv._pback1[t][p][u];
	tsum1+=fbv._pback1[t][p][u];
      }
      cout << endl;
      cout << "_x "<< _x[t] <<" fbv._pback2["<< t<<"][" <<p<<"] "; 
      for (int u=0; u<_nS; u++) {
	cout << " " << fbv._pback2[t][p][u];
	tsum2+=fbv._pback2[t][p][u];
      }
      cout << endl;
      cout << "tsum1 " << tsum1 << " tsum2 " << tsum2 << endl;
    }
  }
 
  for (t=lalgo-2; t>=0; t--) {
    if (t+imin>=irecmin && t+imin<=irecmax) {
      _posteriorM1u[t+imin]=0;
      _posteriorM1d[t+imin]=0;
      _posteriorM2[t+imin]=0;
      _posteriorM3[t+imin]=0;
    }
    double tsum=0;
    for (int p=0; p<_nR; p++) {
      //cout << "*********p=" << p << endl;
      double sumBf = 0;
      for (int m=0; m<5; m++) {tmp_count_alpha[m]=0;}

      for (int u=0; u<_nS; u++) {
	if (fbv._pforw1[t+1][p][u]>eps) {
	  B[u] = fbv._pback2[t+1][p][u]/fbv._pforw1[t+1][p][u];
	} else {
	  B[u] = 0;
	}
	sumBf += B[u]*par._fstat[u];
	Bfcum[u] = sumBf;
      }
      //cout << "sumB " << sumB << endl;
    
      IB[_nS-1] = 0;
      for (int v=_nS-2; v>=0; v--) {
	if (v+1==_nS-1) {
	  IB[v]=par._lambda_i[p]*IB[v+1] + B[v+1];
	} else {
	  IB[v]=par._lambda_i[p]*IB[v+1] + (1-par._lambda_i[p])*B[v+1];
	}
      }
      
      DB[0] = 0;
      for (int v=1; v<_nS; v++) {
	if (v==1) {
	  DB[v]=par._lambda_d[p]*DB[v-1] + B[v-1];
	} else {
	  DB[v]=par._lambda_d[p]*DB[v-1] + (1-par._lambda_d[p])*B[v-1];
	}
      }

      for (int u=0; u<_nS; u++) {
	double an = fbv._pforw3[t][p][u]*par._alpha[p][0]*B[u];
	//double af = fbv._pforw3[t][p][u]*(par._alpha[p][1]/(float)(_nS))*sumB;
	double afdown = fbv._pforw3[t][p][u]*par._alpha[p][1]*Bfcum[u];
	double afup = fbv._pforw3[t][p][u]*par._alpha[p][1]*(sumBf-Bfcum[u]);
	double ai = fbv._pforw3[t][p][u]*par._alpha[p][2]*IB[u];
	double ad = fbv._pforw3[t][p][u]*par._alpha[p][3]*DB[u];
	fbv._pback1[t][p][u] = an+afdown+afup+ai+ad;
	if (u==_nS-1) {
	  double asuppi = fbv._pforw3[t][p][u]*par._alpha[p][2]*B[u];
	  fbv._pback1[t][p][u] += asuppi;
	  tmp_count_alpha[3]+=asuppi;	  
	} 
	if (u==0) {
	  double asuppd = fbv._pforw3[t][p][u]*par._alpha[p][3]*B[u];
	  fbv._pback1[t][p][u] += asuppd;
	  tmp_count_alpha[4]+=asuppd;	  
	} 
	tmp_count_alpha[0]+=an; 
	tmp_count_alpha[1]+=afup;
	tmp_count_alpha[2]+=afdown;
	tmp_count_alpha[3]+=ai;
	tmp_count_alpha[4]+=ad;
	tsum+=fbv._pback1[t][p][u];
      }

      if (t+imin>=irecmin && t+imin<=irecmax) {
	double probi1=fbv._pforw3[t][p][0];
	for (int u=1; u<_nS-1; u++) {
	  probi1 = par._lambda_i[p]*probi1 + fbv._pforw3[t][p][u];
	}
	probi1 += fbv._pforw3[t][p][_nS-1];
	
	par._count_i1[p] += tmp_count_alpha[3] - par._alpha[p][2]*probi1*B[_nS-1];
	
	double probd1=fbv._pforw3[t][p][_nS-1];
	for (int u=_nS-2; u>0; u--) {
	  probd1 = par._lambda_d[p]*probd1 + fbv._pforw3[t][p][u];
	}
	probd1 += fbv._pforw3[t][p][0];
	
	par._count_d1[p] += tmp_count_alpha[4] - par._alpha[p][3]*probd1*B[0];

	_posteriorM1u[t+imin] += tmp_count_alpha[1];
	_posteriorM1d[t+imin] += tmp_count_alpha[2];
	_posteriorM2[t+imin] += tmp_count_alpha[3];
	_posteriorM3[t+imin] += tmp_count_alpha[4];

	par._count_alpha[p][0] += tmp_count_alpha[0];
	par._count_alpha[p][1] += (tmp_count_alpha[1]+tmp_count_alpha[2]);
	par._count_alpha[p][2] += tmp_count_alpha[3];
	par._count_alpha[p][3] += tmp_count_alpha[4];
	
	IC[_nS-1]=0;
	ID[_nS-1]=0;
	IC[_nS-2]=0;
	ID[_nS-2]=0;
	for (int u=_nS-3; u>=0; u--) {
	  double a = (1-par._lambda_i[p]);
	  if (u+2==_nS-1) {a=1;}
	  IC[u]=par._lambda_i[p]*a*B[u+2] + par._lambda_i[p]*(IC[u+1]+ID[u+1]);
	  ID[u]=par._lambda_i[p]*a*B[u+2] + par._lambda_i[p]*ID[u+1];
	}
	double sumi12=0;
	for (int u=0; u<_nS; u++) {
	  sumi12 += fbv._pforw3[t][p][u]*IC[u];
	}
	par._count_i12[p] += par._alpha[p][2]*sumi12;
	
	DC[0]=0;
	DD[0]=0;
	DC[1]=0;
	DD[1]=0;
	for (int u=2; u<_nS; u++) {
	  double a = (1-par._lambda_d[p]);
	  if (u-2==0) {a=1;}
	  DC[u]=par._lambda_d[p]*a*B[u-2] + par._lambda_d[p]*(DC[u-1]+DD[u-1]);
	  DD[u]=par._lambda_d[p]*a*B[u-2] + par._lambda_d[p]*DD[u-1];
	}
	double sumd12=0;
	for (int u=0; u<_nS; u++) {
	  sumd12 += fbv._pforw3[t][p][u]*DC[u];
	}
	par._count_d12[p] += par._alpha[p][3]*sumd12;
      }
    }

    for (int p=0; p<_nR; p++) {
      for (int u=0; u<_nS; u++) {
	fbv._pback1[t][p][u] /= tsum;
      }
    }

    for (int p=0; p<_nR; p++) {
      for (int v=0; v<_nS; v++) {
	fbv._pback2[t][p][v]=0;
	for (int q=0; q<_nR; q++) {
	  fbv._pback2[t][p][v]+=(fbv._pforw2[t][p][v]*par._b[p][q]*fbv._pback1[t][q][v])/fbv._pforw3[t][q][v];
	} 
	if (isnan(fbv._pback2[t][p][v])) {
	  fbv._pback2[t][p][v]=0;
	}
      } 
    }
    
    if (0 && t>imax-10) { //&& (t+imin)%1000==1) {
      for (int p=0; p<_nR; p++) {
	double tsum1=0;
	double tsum2=0;
	cout << "_x "<< _x[t+imin] <<" fbv._pback1["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pback1[t][p][u];
	  tsum1+=fbv._pback1[t][p][u];
	}
	cout << endl;
	cout << "_x "<< _x[t+imin] <<" fbv._pback2["<< t+imin<<"][" <<p<<"] "; 
	for (int u=0; u<_nS; u++) {
	  cout << " " << fbv._pback2[t][p][u];
	  tsum2+=fbv._pback2[t][p][u];
	}
	cout << endl;
	cout << "tsum1 " << tsum1 << " tsum2 " << tsum2 << endl;
      }
    }
    
    if (t+imin>=irecmin && t+imin<=irecmax) {
      this->CIpos(t+imin, fbv._pback1[t], par); 
      this->count_t(par,t+imin);
      this->count_sum_sd(par,t+imin, fbv._pback1[t], fbv._pobs_gaussian[t]);
      for (int p=0; p<_nR; p++) {
	for (int q=0; q<_nR; q++) {
	  double tmp_count_bpq=0;
	  for (int u=0; u<_nS; u++) {
	    tmp_count_bpq+=fbv._pforw1[t][p][u]*par._b[p][q]*(fbv._pback1[t][q][u]/fbv._pforw4[t][q][u]);
	  }
	  if (0 && t==0) {
	    cout << "t " << t << " tmp_count_bpq " << tmp_count_bpq << endl;
	    for (int u=0; u<_nS; u++) {
	      cout << fbv._pforw1[t][p][u] << " " << par._b[p][q] << " " << fbv._pback1[t][q][u] << " " << fbv._pforw4[t][q][u] << endl;
	    }
	  }
	  par._count_bpq[p][q]+=tmp_count_bpq;
	  if (0 && t+imin< 10) {
	    cout << "p=" << p << " q=" << q << " t="<< t+imin << " count_bpq=" << par._count_bpq[p][q] << endl;
	    cout << "comp pforw1";
	    for (int u=0; u<_nS; u++) {
	      cout << " " << fbv._pforw1[t][0][u] - fbv._pforw1[t][1][u]; 
	    }
	    cout << endl;
	    cout << "comp pforw2";
	    for (int u=0; u<_nS; u++) {
	      cout << " " << fbv._pforw2[t][0][u] - fbv._pforw2[t][1][u]; 
	    }
	    cout << endl;
	    cout << "comp pforw3";
	    for (int u=0; u<_nS; u++) {
	      cout << " " << fbv._pforw3[t][0][u] - fbv._pforw3[t][1][u]; 
	    }
	    cout << endl;
	    cout << "comp pforw4";
	    for (int u=0; u<_nS; u++) {
	      cout << " " << fbv._pforw4[t][0][u] - fbv._pforw4[t][1][u]; 
	    }
	    cout << endl;
	    cout << "comp pback1";
	    for (int u=0; u<_nS; u++) {
	      cout << " " << fbv._pback1[t][0][u] - fbv._pback1[t][1][u]; 
	    }
	    cout << endl;
	    cout << "comp pback2";
	    for (int u=0; u<_nS; u++) {
	      cout << " " << fbv._pback2[t][0][u] - fbv._pback2[t][1][u]; 
	    }
	    cout << endl;
	  }
	}
      }
    }
  }

  delete IB;
  delete DB;
  delete IC;
  delete ID;
  delete DC;
  delete DD;

  delete B;
  delete tmp_count_alpha;
}

Data::~Data() 
{
}


