/* Copyright (C) 2009 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "params.h"
#include "data.h"

#include<iostream>
#include<vector>
#include<cmath>
#include<gsl/gsl_multimin.h>
#include<gsl/gsl_rng.h>
#include<gsl/gsl_randist.h>
#include<gsl/gsl_cdf.h>
#include<gsl/gsl_sort_vector.h>

using namespace std;

Params::Params(int nS, int nR, float xmin, float xmax, float smin, float smax, float* alphamin, float maxdriftamplitude)
{
  _nS = nS;
  _nR = nR;
  _xmin = xmin;
  _xmax = xmax;

   _mu = new float [_nS];
   for (int i=0; i<_nS; i++) {
     _mu[i] = smin + i*(smax-smin)/(float)(_nS-1);
   }
   cout << "_mus " << _mu[0] << " " << _mu[1] << " ... " << _mu[_nS-1] << endl;

   _alphamin = new float [4];
   for (int ia=0; ia<4; ia++) {
     _alphamin[ia]=alphamin[ia];
   }
   _maxdriftamplitude = maxdriftamplitude;
   this->compute_maxlambda();
   
   _alpha = new float* [_nR];
   _b = new float* [_nR];
   _lambda_i = new float [_nR];
   _lambda_d = new float [_nR];
   for (int p=0; p<_nR; p++) {
     _alpha[p] = new float [4]; // m=n,f,i,d
     _b[p] = new float [_nR];
   }   

   _r0 = new float [_nR];
   _s0 = new float [_nS];

   _count_alpha = new double* [_nR];
   for (int p=0; p<_nR; p++) {
     _count_alpha[p] = new double [4]; 
   } 

   _count_i12 = new double [_nR];
   _count_i1 = new double [_nR];
   _count_d12 = new double [_nR];
   _count_d1 = new double [_nR];
   _pobs_outlier = new float [4];

   _pobs_outlier[0] = 1.0;
   _pobs_outlier[1] = 1.0/(smax-smin);
   _pobs_outlier[2] = (xmin-smin)/(smax-smin);
   _pobs_outlier[3] = (smax-xmax)/(smax-smin);

   _rho_v = new float [_nS];
   _sigma_v = new float [_nS];
   _sigma2_v = new float [_nS];

   _count_sum_rhosd_0 = new double [_nS];
   _count_sum_rhosd_1 = new double [_nS];
   _count_sum_rhosd_2 = new double [_nS];
   _count_sum_rhosd_3 = new double [_nS];


   _count_bpq = new double* [_nR];
   for (int p=0; p<_nR; p++) {
     _count_bpq[p] = new double [_nR]; 
   } 

   _type_alpha = new int [_nR];
   _type_lambda = new int [_nR];
   _cat_mu = NULL;
   _type_sigma = -1;
   _type_rho = -1;
   _type_outlier = -1;
   _type_b = -1;
   for (int r=0; r<_nR; r++) {
     _type_alpha[r]=-1;
     _type_lambda[r]=-1;
   } 
   _fstat = new float [_nS];
}


void Params::alloclogit(int ncatlogit)
{
  _ncatlogit = ncatlogit;
  _poutlierlogit = new double [_ncatlogit];
  if (_type_outlier == 2) {
    _outlier_logit = gsl_vector_alloc(2);
    _count_logitoutlier = new double* [_ncatlogit];
    for (int i=0; i<_ncatlogit; i++) {
      _poutlierlogit[i] = 1/(1+exp(-(_outlier_logita + _outlier_logitb*((float)i)/100.0)));
      _count_logitoutlier[i] = new double [2];
    }
  } else {
    for (int i=0; i<_ncatlogit; i++) {
      _poutlierlogit[i] = _poutlier;
    }
  }
}

void Params::definecatmu(double frac_cat_mu, Data** datav, int ndata)
{
  //cout << "now definecatmu" <<endl;
  vector<int> cat_mu;
  int ntot = 0;
  for (int id=0; id<ndata; id++) {
    ntot += datav[id]->_lseq;
  }

  gsl_vector *alldata;
  alldata = gsl_vector_alloc(ntot);
  int k=0;
  for (int id=0; id<ndata; id++) {
    for (int t=0; t<datav[id]->_lseq; t++) {
      gsl_vector_set(alldata,k,datav[id]->_x[t]);
      k++;
    }      
  }  
  gsl_sort_vector(alldata);
  int lastbreak=0;
  int n=0;
  int nold=0;
  for (float p=frac_cat_mu; p<1; p+=frac_cat_mu) {
    float x = gsl_vector_get(alldata,(int)(p*ntot));
    //cout << " p " << p << " x " << x << endl;
    int bestu = _nS-1;
    for (int u=lastbreak+1; u<_nS; u++) {
      if (fabs(x-(_mu[bestu]+_mu[bestu-1])/2.0)>fabs(x-(_mu[u]+_mu[u-1])/2.0)) {
	int n = 0;
	for (int id=0; id<ndata; id++) {
	  n += datav[id]->nxinfmu((_mu[u]+_mu[u-1])/2);
	}
	double fr=(double)(n-nold)/(double)(ntot);
	//cout << " u="<< u << " fr="<< fr << endl;
	if (fr>0.9*frac_cat_mu) {
	  bestu = u;
	}
      }
    }
    if (bestu<_nS && bestu>lastbreak) {
      lastbreak=bestu;
      nold=n;
      //cout << bestu << " " << fabs(x-_mu[bestu]) << endl;
      cat_mu.push_back(bestu);
    }
  }
  cat_mu.push_back(_nS);
  gsl_vector_free(alldata);

//   double px=0;
//   for (int u=0; u<_nS-1; u++) {
//     int n = 0;
//     for (int id=0; id<ndata; id++) {
//       n += datav[id]->nxinfmu((_mu[u]+_mu[u+1])/2);
//     }
//     double p = (double)n/(double)ntot;
//     if (p-px>=frac_cat_mu && 1-p>frac_cat_mu/2.0) {
//       cat_mu.push_back(u+1);
//       px=p;
//     }
//   }
//   cat_mu.push_back(_nS);
  _ncat_mu = cat_mu.size();
  if (_cat_mu != NULL) {
    delete _cat_mu;
  }
  _cat_mu = new int [_ncat_mu];
  for (int i=0; i<_ncat_mu; i++) {
    _cat_mu[i] = cat_mu[i];
    //cout << "_cat_mu[" << i << "] " << _cat_mu[i] << endl;
  }
}

void Params::definefstatkdens(Data** datav, int ndata, bool etaflat)
{
  if (etaflat) {
    for (int u=0; u<_nS; u++) {
      _fstat[u]=1.0/(double)_nS;
    }
  } else {
    int ntot = 0;
    for (int id=0; id<ndata; id++) {
      for (int t=0; t<datav[id]->_lseq; t++) {
	if (datav[id]->_obs_status!=0) {
	  ntot++;
	}
      }
    }
    
    gsl_vector *alldata;
    alldata = gsl_vector_alloc(ntot);
    int k=0;
    for (int id=0; id<ndata; id++) {
      for (int t=0; t<datav[id]->_lseq; t++) {
	if (datav[id]->_obs_status!=0) {
	  gsl_vector_set(alldata,k,datav[id]->_x[t]);
	  k++;
	}
      }      
    }  
    gsl_sort_vector(alldata);
    
    double quartilediff = gsl_vector_get(alldata,(int)(0.75*ntot))-gsl_vector_get(alldata,(int)(0.25*ntot));
    
    double sumx=0;
    double sumx2=0;
    for (int i=0; i<ntot; i++) {
      sumx+=gsl_vector_get(alldata,i);
      sumx2+=pow(gsl_vector_get(alldata,i),2);
    }
    double sdx = sqrt(sumx2/ntot-pow(sumx/ntot,2));
    double minnum=sdx;
    if (quartilediff/1.34<minnum) { minnum=quartilediff/1.34; }
    double bwscott=1.06*(minnum)*pow(ntot,-1./5.);
    cout << "bwscott " << bwscott << endl;
    float h =_mu[1]-_mu[0];
    float sum=0;
    for (int u=0; u<_nS; u++) {
      float xdo=_mu[u]-h/2.0;
      float xup=_mu[u]+h/2.0;
      _fstat[u]=(1./(ntot+1))*(h/(_mu[_nS-1]-_mu[0]+h)); // pseudocount
      for (int i=0; i<ntot; i++) {
	double xobs=gsl_vector_get(alldata,i);
	_fstat[u] += (1./(ntot+1))*(gsl_cdf_gaussian_P(xup-xobs,bwscott)-gsl_cdf_gaussian_P(xdo-xobs,bwscott));
      }
      sum+=_fstat[u];
    }
    //cout << "sum " << sum << endl;
    for (int u=0; u<_nS; u++) {
      _fstat[u]/=sum;
    }
    gsl_vector_free(alldata);
  }
}

void Params::copycatmufstat(Params &templ)
{
  _ncat_mu = templ._ncat_mu;
  _cat_mu = new int [_ncat_mu];
  for (int i=0; i<_ncat_mu; i++) {
    _cat_mu[i] = templ._cat_mu[i];
  }
  for (int u=0; u<_nS; u++) {
    _fstat[u] = templ._fstat[u];
  }
}


void Params::initdeterministic(int type_sigma, int type_rho, int type_outlier)
{
  _type_sigma = type_sigma;
  _type_rho = type_rho;
  _type_outlier = type_outlier;
  _type_b = 1;
  for (int u=0; u<_nS; u++) {
    _sigma_v[u]=1;
    _sigma2_v[u]=pow(_sigma_v[u],2);
    _rho_v[u]=1;
  }

  if (type_outlier==1) {
    _poutlier=0.05;
  } else if (type_outlier==2) {
    _outlier_logita = -log(19);
    _outlier_logitb = 0;
  }

  for (int p=0; p<_nR; p++) {
    _type_alpha[p] = 1;
    _type_lambda[p] = 1;
    _r0[p] = 1.0/(float)(_nR);
    _lambda_i[p]=0.8;
    _lambda_d[p]=0.8;
    this->force_maxlambda(p);

    _alpha[p][0] = 0.9-0.1*(double)p; // m=n,f,i,d
    _alpha[p][1] = (1-_alpha[p][0])/3;
    _alpha[p][2] = (1-_alpha[p][0])/3;
    _alpha[p][3] = (1-_alpha[p][0])/3;
    
    _b[p][p] = 1 - (_nR-1)*0.01;
    for (int q=0; q<_nR; q++) {
      if (q!=p) {
	_b[p][q]=0.01;
      }
    }
  }    
  for (int u=0; u<_nS; u++) {
    _s0[u] = 1.0/(float)(_nS);
  }   
}

void Params::initrandom(gsl_rng * rng)
{
  int ok=1;
  if (_type_sigma<0 || _type_rho<0 || _type_outlier<0 || _type_b<0) {
    ok=0;
  }
  for (int r=0; r<_nR; r++) {
    if (_type_alpha[r]<0 || _type_lambda[r]<0) {
      ok=0;
    }
  }
  if (ok!=1) {
    cerr << "pb in void Params::initrandom(gsl_rng * r)" << endl;
    exit(0);
  }

  if (_type_sigma!=0) {
    float sigma = gsl_ran_exponential(rng, 1);
    for (int u=0; u<_nS; u++) {
      _sigma_v[u]=sigma;
      _sigma2_v[u]=pow(_sigma_v[u],2);
    }
  }
  if (_type_rho!=0) {
    float rho = gsl_ran_beta(rng, 1, 1);
    for (int u=0; u<_nS; u++) {
      _rho_v[u]=rho;
    }
  }

  if (_type_outlier!=0) {
    if (_type_outlier==1) {
      _poutlier=gsl_ran_beta(rng, 1, 19);
    } else if (_type_outlier==2) {
      float poutl = gsl_ran_beta(rng, 1, 19);
      _outlier_logita = log(poutl);
      _outlier_logitb = 0;
    }
  }

  if (_type_b!=0) {
    for (int p=0; p<_nR; p++) {
      _b[p][p] = 1 - (_nR-1)*0.01;
      for (int q=0; q<_nR; q++) {
	if (q!=p) {
	  _b[p][q]=0.01;
	}
      }
    }
  }
  for (int p=0; p<_nR; p++) {
    _r0[p] = 1.0/(float)(_nR);
    if (_type_alpha[p]!=0) {
      _alpha[p][0] = gsl_ran_exponential(rng, 1);
      _alpha[p][1] = gsl_ran_exponential(rng, 0.01);
      if (_type_alpha[p]<2) {
	_alpha[p][2] = gsl_ran_exponential(rng, 0.1);
	_alpha[p][3] = gsl_ran_exponential(rng, 0.1);
      } else if (_type_alpha[p]>=2) { 
	// Useful in order to set this two parameters to zero
	if (_type_alpha[p]==3) {
	  _alpha[p][2] = gsl_ran_exponential(rng, 0.1);
	}
	if (_type_alpha[p]==4) {	
	  _alpha[p][3] = gsl_ran_exponential(rng, 0.1);
	}
      }
      float tsum = _alpha[p][0]+_alpha[p][1]+_alpha[p][2]+_alpha[p][3];
      for (int m=0; m<4; m++) {
	_alpha[p][m] /= tsum;
      }
      this->force_alphamin(p);
    }
    if ( _type_lambda[p]!=0) {
      _lambda_i[p]=gsl_ran_flat(rng, 0.3, 1);
      _lambda_d[p]=_lambda_i[p]; //gsl_ran_beta(rng, 1, 1);
      this->force_maxlambda(p);
    } 
  }  
  
  for (int u=0; u<_nS; u++) {
    _s0[u] = 1.0/(float)(_nS);
  }   
}


void Params::initrandom(int type_sigma, int type_rho, int type_outlier, gsl_rng * r)
{
  _type_sigma = type_sigma;
  _type_rho = type_rho;
  _type_outlier = type_outlier;
  _type_b = 1;


  float sigma = gsl_ran_exponential(r, 1);
  float rho = 0;
  if (_type_rho!=0) {
    rho = gsl_ran_beta(r, 1, 1);    
  }
  
  for (int u=0; u<_nS; u++) {
    _sigma_v[u]=sigma;
    _sigma2_v[u]=pow(_sigma_v[u],2);
    _rho_v[u]=rho;
  }

  if (type_outlier==1) {
    _poutlier=gsl_ran_beta(r, 1, 19);
  } else if (type_outlier==2) {
    float poutl = gsl_ran_beta(r, 1, 19);
    _outlier_logita = log(poutl);
    _outlier_logitb = 0;
  }

   for (int p=0; p<_nR; p++) {
     _type_alpha[p] = 1;
     _type_lambda[p] = 1;
     _r0[p] = 1.0/(float)(_nR);
     _lambda_i[p]=gsl_ran_flat(r, 0.3, 1);
     _lambda_d[p]=_lambda_i[p]; //gsl_ran_beta(r, 1, 1);
     this->force_maxlambda(p);

     float tsum = 0;
//      for (int m=0; m<4; m++) {
//        _alpha[p][m] = gsl_ran_beta(r, 1, 1); // m=n,f,i,d
//        tsum += _alpha[p][m];
//      }
     _alpha[p][0] = gsl_ran_exponential(r, 1);
     _alpha[p][1] = gsl_ran_exponential(r, 0.01);
     _alpha[p][2] = gsl_ran_exponential(r, 0.1);
     _alpha[p][3] = gsl_ran_exponential(r, 0.1);
     tsum = _alpha[p][0]+_alpha[p][1]+_alpha[p][2]+_alpha[p][3];
     for (int m=0; m<4; m++) {
       _alpha[p][m] /= tsum;
     }
      this->force_alphamin(p);
     _b[p][p] = 1 - (_nR-1)*0.01;
     for (int q=0; q<_nR; q++) {
       if (q!=p) {
	 _b[p][q]=0.01;
       }
     }
   }    
   for (int u=0; u<_nS; u++) {
     _s0[u] = 1.0/(float)(_nS);
   }   
}


Params::Params(ifstream& in)
{
  string word;

  in >> word;
  if (word.compare("nS")!=0) {
    cout << "Error in parameter input file : keyword \"nS\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _nS;

  in >> word;
  if (word.compare("mu")!=0) {
    cout << "Error in parameter input file : keyword \"mu\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  _mu = new float [_nS];
  for (int u=0; u<_nS; u++) {
    in >> _mu[u];
  }

  in >> word;
  if (word.compare("eta")!=0) {
    cout << "Error in parameter input file : keyword \"eta\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  _fstat = new float [_nS];
  for (int u=0; u<_nS; u++) {
    in >> _fstat[u];
  }

  in >> word;
  if (word.compare("xmin")!=0) {
    cout << "Error in parameter input file : keyword \"xmin\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _xmin;

  in >> word;
  if (word.compare("xmax")!=0) {
    cout << "Error in parameter input file : keyword \"xmax\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _xmax;

  in >> word;
  if (word.compare("ncat_mu")!=0) {
    cout << "Error in parameter input file : keyword \"ncat_mu\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _ncat_mu;
  _cat_mu = new int [_ncat_mu];

  in >> word;
  if (word.compare("cat_mu")!=0) {
    cout << "Error in parameter input file : keyword \"cat_mu\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  for (int ic=0; ic<_ncat_mu; ic++) {
    in >> _cat_mu[ic];
  }

  in >> word;
  if (word.compare("sigma_type")!=0) {
    cout << "Error in parameter input file : keyword \"sigma_type\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _type_sigma;

  in >> word;
  if (word.compare("sigma_val")!=0) {
    cout << "Error in parameter input file : keyword \"sigma_val\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  _sigma_v = new float [_nS];
  for (int u=0; u<_nS; u++) {
    in >> _sigma_v[u];
  }    

  in >> word;
  if (word.compare("rho_type")!=0) {
    cout << "Error in parameter input file : keyword \"rho_type\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _type_rho;

  in >> word;
  if (word.compare("rho_val")!=0) {
    cout << "Error in parameter input file : keyword \"rho_val\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  _rho_v = new float [_nS];
  for (int u=0; u<_nS; u++) {
    in >> _rho_v[u];
  }    

  in >> word;
  if (word.compare("outlier_type")!=0) {
    cout << "Error in parameter input file : keyword \"outlier_type\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _type_outlier;

  if (_type_outlier == 0 || _type_outlier == 1) {
    in >> word;
    if (word.compare("outlier")!=0) {
      cout << "Error in parameter input file : keyword \"outlier\" expected instead of \"" << word << "\""<< endl;
      exit(1);
    }
    in >> _poutlier;
  } else if (_type_outlier == 2) {
    in >> word;
    if (word.compare("outlier_logit")!=0) {
      cout << "Error in parameter input file : keyword \"outlier_logit\" expected instead of \"" << word << "\""<< endl;
      exit(1);
    }
    in >> _outlier_logita;    
    in >> _outlier_logitb;    
  }

  in >> word;
  if (word.compare("nR")!=0) {
    cout << "Error in parameter input file : keyword \"nR\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _nR;

  in >> word;
  if (word.compare("b_type")!=0) {
    cout << "Error in parameter input file : keyword \"b_type\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  in >> _type_b;

  in >> word;
  if (word.compare("b")!=0) {
    cout << "Error in parameter input file : keyword \"b\" expected instead of \"" << word << "\""<< endl;
    exit(1);
  }
  _b = new float* [_nR];
  for (int p=0; p<_nR; p++) {
    _b[p] = new float [_nR];    
    for (int q=0; q<_nR; q++) {
      in >> _b[p][q];
    }    
  }

  _type_alpha = new int [_nR];
  _type_lambda = new int [_nR];
  _alpha = new float* [_nR];
  _lambda_i = new float [_nR];
  _lambda_d = new float [_nR];
  for (int p=0; p<_nR; p++) {
    _alpha[p] = new float [4];
  }

  for (int p=0; p<_nR; p++) {
    in >> word;
    if (word.compare("alpha_type")!=0) {
      cout << "Error in parameter input file : keyword \"alpha_type\" expected instead of \"" << word << "\""<< endl;
      exit(1);
    }
    in >> _type_alpha[p];

    in >> word;
    if (word.compare("alpha")!=0) {
      cout << "Error in parameter input file : keyword \"alpha\" expected instead of \"" << word << "\""<< endl;
      exit(1);
    }
    for (int k=0; k<4; k++) {
      in >> _alpha[p][k];
    }
    
    in >> word;
    if (word.compare("lambda_type")!=0) {
      cout << "Error in parameter input file : keyword \"lambda_type\" expected instead of \"" << word << "\""<< endl;
      exit(1);
    }
    in >> _type_lambda[p];
    in >> word;
    if (word.compare("lambda")!=0) {
      cout << "Error in parameter input file : keyword \"lambda\" expected instead of \"" << word << "\""<< endl;
      exit(1);
    }
    in >> _lambda_i[p];
    in >> _lambda_d[p];
  }

  _alphamin = new float [4];
  for (int ia=0; ia<4; ia++) {
    _alphamin[ia]=0;
  }
  in >> word;
  if (word.compare("alphamin")==0) {
    cout << "keyword \"alphamin\" found, reading alphamin values:";
    for (int ia=0; ia<4; ia++) {
      in >> _alphamin[ia];
      cout << " " << _alphamin[ia];
    }
    cout << endl;
  }
  _maxdriftamplitude=-1;
  if (word.compare("maxdriftamplitude")==0) {
    in >> _maxdriftamplitude;
    cout << "keyword \"maxdriftamplitude\" read: " << _maxdriftamplitude << endl;
  }
  this->compute_maxlambda();
  
  _pobs_outlier = new float [4];
  
  _pobs_outlier[0] = 1.0;
  _pobs_outlier[1] = 1.0/(_mu[_nS-1]-_mu[0]);
  _pobs_outlier[2] = (_xmin-_mu[0])/(_mu[_nS-1]-_mu[0]);
  _pobs_outlier[3] = (_mu[_nS-1]-_xmax)/(_mu[_nS-1]-_mu[0]);

   _r0 = new float [_nR];
   _s0 = new float [_nS];

   _count_alpha = new double* [_nR];
   for (int p=0; p<_nR; p++) {
     _count_alpha[p] = new double [4]; 
   } 

   _count_i12 = new double [_nR];
   _count_i1 = new double [_nR];
   _count_d12 = new double [_nR];
   _count_d1 = new double [_nR];

   _sigma2_v = new float [_nS];
   for (int u=0; u<_nS; u++) {
     _sigma2_v[u] = pow(_sigma_v[u],2);
   }

   _count_sum_rhosd_0 = new double [_nS];
   _count_sum_rhosd_1 = new double [_nS];
   _count_sum_rhosd_2 = new double [_nS];
   _count_sum_rhosd_3 = new double [_nS];


   _count_bpq = new double* [_nR];
   for (int p=0; p<_nR; p++) {
     _count_bpq[p] = new double [_nR]; 
   } 
  
   for (int p=0; p<_nR; p++) {
     _r0[p] = 1.0/(float)(_nR);
   }    
   for (int u=0; u<_nS; u++) {
     _s0[u] = 1.0/(float)(_nS);
   }   

   for (int p=0; p<_nR; p++) {
     double asum=0;
     for (int im=0; im<4; im++) {
       asum += _alpha[p][im];
     }
     for (int im=0; im<4; im++) {
       _alpha[p][im] /= asum;
     }
   } 

   for (int p=0; p<_nR; p++) {
     double bsum=0;
     for (int q=0; q<_nR; q++) {
       bsum += _b[p][q];
     }
     for (int q=0; q<_nR; q++) {
       _b[p][q] /= bsum;
     }
     
   } 
   
}


void Params::print(ofstream& out)
{
  //out << "#general" << endl;
  out << "nS " << _nS << endl;
  out << "mu";
  for (int u=0; u<_nS; u++) {
    out << " " << _mu[u];
  }
  out << endl;
  out << "eta";
  for (int u=0; u<_nS; u++) {
    out << " " << _fstat[u];
  }
  out << endl;

  //out << "#observations" << endl;
  out << "xmin " << _xmin << " xmax " << _xmax << endl;
  out << "ncat_mu " << _ncat_mu << endl;
  out << "cat_mu";
  for (int ic=0; ic<_ncat_mu; ic++) {
    out << " " << _cat_mu[ic];
  }
  out << endl;
  
  out << "sigma_type " << _type_sigma << endl;
  out << "sigma_val";
  for (int u=0; u<_nS; u++) {
    out << " " << _sigma_v[u];
  }    
  out << endl;

  out << "rho_type " << _type_rho << endl;
  out << "rho_val";
  for (int u=0; u<_nS; u++) {
    out << " " << _rho_v[u];
  }    
  out << endl;
  
  out << "outlier_type " << _type_outlier << endl;
  if (_type_outlier != 2) {
    out << "outlier " << _poutlier << endl;
  } else {
    out << "outlier_logit " << _outlier_logita << " " << _outlier_logitb << endl;
  }

  //out << "#hidden states" << endl;
  out << "nR " << _nR << endl;
  out << "b_type " << _type_b << endl;
  out << "b" << endl;
  for (int p=0; p<_nR; p++) {
    for (int q=0; q<_nR; q++) {
      out << " " << _b[p][q];
    }    
    out << endl;
  }
  for (int p=0; p<_nR; p++) {
    out << "alpha_type " << _type_alpha[p] << endl;
    out << "alpha";
    for (int k=0; k<4; k++) {
      out << " " << _alpha[p][k];
    }
    out << endl;
    out << "lambda_type " << _type_lambda[p] << endl;
    out << "lambda " << _lambda_i[p] << " " << _lambda_d[p] << endl;
  }
  out << "alphamin ";
  for (int ia=0; ia<4; ia++) {
    out << " " << _alphamin[ia];
  }
  out << endl;  
}

void Params::reset_counts()
{
  _count_t=0;
  _count_sum_sd=0;
   for (int p=0; p<_nR; p++) {
     for (int m=0; m<4; m++) {
       _count_alpha[p][m] = 0;
     }
     _count_i12[p]=0;
     _count_i1[p]=0;
     _count_d12[p]=0;
     _count_d1[p]=0;
     for (int q=0; q<_nR; q++) {
       _count_bpq[p][q]=0;
     }
   } 
   _count_outlier = 0;
   for (int u=0; u<_nS; u++) {
     _count_sum_rhosd_0[u]=0;
     _count_sum_rhosd_1[u]=0;
     _count_sum_rhosd_2[u]=0;
     _count_sum_rhosd_3[u]=0;
   }

   if (_type_outlier == 2) {
     for (int c=0; c<_ncatlogit; c++) {
       _count_logitoutlier[c][0]=0;
       _count_logitoutlier[c][1]=0;
     }
   }

}

void Params::compute_maxlambda()
{
  if (_maxdriftamplitude<0) {
    _maxlambda=-1;
  } else {
    _maxlambda = 1 - (_mu[1]-_mu[0])/_maxdriftamplitude;
    if (_maxlambda<0) {
      _maxlambda = 0;
    }
    cout << "_maxlambda set to " << _maxlambda << " (based on grid size and user provided \"maxdriftamplitude\")" << endl;
  }
}

void Params::force_maxlambda(int p)
{
  if (_maxlambda>0) {
    if (_lambda_i[p]>_maxlambda) {
      _lambda_i[p]=_maxlambda;
    }
    if (_lambda_d[p]>_maxlambda) {
      _lambda_d[p]=_maxlambda;
    }
  }
}

void Params::force_alphamin(int p)
{
  if (_type_alpha[p]>=1) {
    cout << "now forcing alphamin" << endl;
    bool alphaminhit[4] = {false, false, false, false};
    int n_newalphaminhit = 0;
    int iteralphaminhit = 0;
    while (iteralphaminhit==0 || n_newalphaminhit>0) {
      iteralphaminhit++;
      n_newalphaminhit=0;
      double tsum=0;
      double tsum_alphamin = 0;
      cout << " init alpha";
      for (int m=0; m<4; m++) {
	cout << " " << _alpha[p][m];
	if (_alpha[p][m]<_alphamin[m] && !alphaminhit[m]) {
	  alphaminhit[m]=true;
	  n_newalphaminhit++;
	}
	if (!alphaminhit[m]) {
	  tsum += _alpha[p][m];
	  //tsum += _count_alpha[p][m];
	} else {
	  tsum_alphamin += _alphamin[m];
	}
      }
      cout << endl;
      cout << " new alpha";
      for (int m=0; m<4; m++) {
	if (!alphaminhit[m]) {
	  _alpha[p][m] = (1-tsum_alphamin)*_alpha[p][m]/tsum;
	  //_alpha[p][m] = (1-tsum_alphamin)*_count_alpha[p][m]/tsum;
	} else {
	  _alpha[p][m] = _alphamin[m];
	}
	cout << " " << _alpha[p][m];
	if (alphaminhit[m]) {
	  cout << "(min)";
	}
      }
      cout << endl;
    }      
  }
}

void Params::m_step(int iter)
{
  if (_type_rho==1) {
    double sum_rho_1 = 0;
    double sum_rho_2 = 0;
    for (int u=0; u<_nS; u++) {
      sum_rho_1 += _count_sum_rhosd_1[u];
      sum_rho_2 += _count_sum_rhosd_2[u];
    }
    
    double rho=sum_rho_2/sum_rho_1;
    if (!isnan(rho) && rho>=0 && rho<=1) {
      for (int u=0; u<_nS; u++) {
	_rho_v[u]=rho;
      }      
    }
  } else if (_type_rho==2) {
    int umin=0;
    //cout << "u _count_sum_rhosd_0[u] _count_sum_rhosd_1[u] _count_sum_rhosd_2[u]_count_sum_rhosd_3[u]" << endl;
    for (int ucat=0; ucat<_ncat_mu; ucat++) {
      double sum_rho_1 = 0;
      double sum_rho_2 = 0;
      for (int u=umin; u<_cat_mu[ucat]; u++) {
	sum_rho_1 += _count_sum_rhosd_1[u];
	sum_rho_2 += _count_sum_rhosd_2[u];
	//cout << u << " " << _count_sum_rhosd_0[u] << " " << _count_sum_rhosd_1[u] << " " << _count_sum_rhosd_2[u] << " " << _count_sum_rhosd_3[u] << endl;
      }
      
      double rho=sum_rho_2/sum_rho_1;
      if (!isnan(rho) && rho>=0 && rho<=1) {
	for (int u=umin; u<_cat_mu[ucat]; u++) {
	  _rho_v[u]=rho;
	}      
      }
      umin = _cat_mu[ucat];      
    }    
  }

  if (_type_sigma==1) {
    double sum_sd_0 = 0;
    double sum_sd_1 = 0;
    double sum_sd_2 = 0;
    double sum_sd_3 = 0;
    for (int u=0; u<_nS; u++) {
      sum_sd_0 += _count_sum_rhosd_0[u];
      sum_sd_1 += pow(_rho_v[u],2)*_count_sum_rhosd_1[u];
      sum_sd_2 += _rho_v[u]*_count_sum_rhosd_2[u];
      sum_sd_3 += _count_sum_rhosd_3[u];
    }
    double sigma2=(sum_sd_3 -2*sum_sd_2 + sum_sd_1)/sum_sd_0;
    double sigma=sqrt(sigma2);
    if (!isnan(sigma2)) {
      for (int u=0; u<_nS; u++) {
	_sigma2_v[u]=sigma2;
	_sigma_v[u]=sigma;	
      }
    }
  } else if (_type_sigma==2) {
    int umin=0;
    for (int ucat=0; ucat<_ncat_mu; ucat++) {
      double sum_sd_0 = 0;
      double sum_sd_1 = 0;
      double sum_sd_2 = 0;
      double sum_sd_3 = 0;
      for (int u=umin; u<_cat_mu[ucat]; u++) {
	sum_sd_0 += _count_sum_rhosd_0[u];
	sum_sd_1 += pow(_rho_v[u],2)*_count_sum_rhosd_1[u];
	sum_sd_2 += _rho_v[u]*_count_sum_rhosd_2[u];
	sum_sd_3 += _count_sum_rhosd_3[u];
      }
      double sigma2=(sum_sd_3 -2*sum_sd_2 + sum_sd_1)/sum_sd_0;
      double sigma=sqrt(sigma2);
      if (!isnan(sigma2)) {
	for (int u=umin; u<_cat_mu[ucat]; u++) {
	  _sigma2_v[u]=sigma2;
	  _sigma_v[u]=sigma;	
	}
      }
      umin = _cat_mu[ucat];
    }    
  }
  
  if (_type_outlier==1) {
    _poutlier = _count_outlier/(double)(_count_t);
    for (int i=0; i<_ncatlogit; i++) {
      _poutlierlogit[i] = _poutlier;
    }
  } else if (_type_outlier==2) {
    float tsum = 0;
    //cout << "_count_logitoutlier";
    for (int c=0; c<_ncatlogit; c++) {
      //cout << " " << 0.01*(double)c << ":" << _count_logitoutlier[c][0] << "+" << _count_logitoutlier[c][1];
      tsum += _count_logitoutlier[c][0] + _count_logitoutlier[c][1]; 
    }
    //cout << endl;
    //cout << "sum " << tsum << endl;
    gsl_vector_set(_outlier_logit,0,_outlier_logita);
    gsl_vector_set(_outlier_logit,1,_outlier_logitb);
    //cout << "logl " << compute_loglik_poutlier(_outlier_logit, this) << endl;
    const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;
    gsl_multimin_fminimizer *s = NULL;
    gsl_multimin_function minex_func;
    gsl_vector* ss = gsl_vector_alloc (2);
    gsl_vector_set_all (ss, 1.0);
    minex_func.n = 2;
    minex_func.f = &compute_loglik_poutlier;
    minex_func.params = (void *)this;
    s = gsl_multimin_fminimizer_alloc (T, 2);
    gsl_multimin_fminimizer_set (s, &minex_func,_outlier_logit, ss);
     
    int status;
    double size;
    do
      {
	iter++;
	status = gsl_multimin_fminimizer_iterate(s);
	
	if (status) 
	  break;
	
	size = gsl_multimin_fminimizer_size (s);
	status = gsl_multimin_test_size (size, 1e-4);
     
	if (status == GSL_SUCCESS)
	  {
	    //printf ("converged to minimum at\n");
	  }
	
// 	printf ("%5d %10.3e %10.3ef f() = %7.3f size = %.3f\n", 
// 		iter,
// 		gsl_vector_get (s->x, 0), 
// 		gsl_vector_get (s->x, 1), 
// 		s->fval, size);
      }
    while (status == GSL_CONTINUE && iter < 100);
       
    _outlier_logita = gsl_vector_get(s->x,0);
    _outlier_logitb = gsl_vector_get(s->x,1);
    //gsl_vector_free(x);
    gsl_vector_free(ss);
    gsl_multimin_fminimizer_free (s);

    for (int i=0; i<_ncatlogit; i++) {
      _poutlierlogit[i] = 1/(1+exp(-(_outlier_logita + _outlier_logitb*((float)i)/100.0)));
    }
  }


  

  for (int p=0; p<_nR; p++) {
    if (_type_alpha[p]>=1) {
      double tsum=0;
      for (int m=0; m<4; m++) {
	tsum += _count_alpha[p][m];
      }
      
      for (int m=0; m<4; m++) {
	_alpha[p][m] = _count_alpha[p][m]/tsum;
      }
      this->force_alphamin(p);
    }
    
    if (_type_lambda[p]==1) {
      float tli = _count_i12[p]/(_count_i12[p]+_count_i1[p]);
      if (isfinite(tli)) {
	_lambda_i[p] = tli;    
      }
      float tld = _count_d12[p]/(_count_d12[p]+_count_d1[p]);
      if (isfinite(tld)) {
	_lambda_d[p] = tld;
      }
      this->force_maxlambda(p);
    }
  }
  
  if (_type_b==1) {
    for (int p=0; p<_nR; p++) {
      double tsum=0;
      for (int q=0; q<_nR; q++) {
	tsum += _count_bpq[p][q];
	//cout << "_count_bpq[p][q] " << _count_bpq[p][q] << endl;
      }
      for (int q=0; q<_nR; q++) {
	if (1 || iter<4) {
	  _b[p][q] = _count_bpq[p][q]/tsum;
	}
      }
    }  
  }
  
}

Params::~Params()
{
}

double compute_loglik_poutlier(const gsl_vector *x, void *params)
{
  Params* par = (Params*)params;
  double logl = 0;
  //cout << "params " << gsl_vector_get(x,0) << " " << gsl_vector_get(x,1)<< endl;
  for (int c=0; c<par->_ncatlogit; c++) {
    if (par->_count_logitoutlier[c][0]+par->_count_logitoutlier[c][1]>1e-30) {
      //cout << " " << par->_count_logitoutlier[c][0]+par->_count_logitoutlier[c][1];
      double p=1/(1+exp(-(gsl_vector_get(x,0) + 0.01*gsl_vector_get(x,1)*(double)c)));
      logl += par->_count_logitoutlier[c][0]*log(p)+ par->_count_logitoutlier[c][1]*log(1-p);
    }
  }
  //cout << endl;
  //cout << "computed logl " << logl << endl;
  return(-logl);
}
