/* Copyright (C) 2009, 2012, 2021 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _PARAMS_H_
#define _PARAMS_H_

#include <fstream>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_vector.h>

using namespace std;

class Data;

class Params
{

 public:
  int _nS;
  int _nR;
  float _xmin;
  float _xmax;
  float* _mu; // length nS
  float* _alphamin;
  float _maxdriftamplitude;
  float _maxlambda;
  
  float* _rho_v; //length nS
  float* _sigma2_v;
  float* _sigma_v;
  int _ncat_mu;
  int* _cat_mu;
  
  int _type_sigma;
  int _type_rho;
  int _type_outlier;
  int _type_b;
  int* _type_alpha;
  int* _type_lambda;

  float** _alpha; // _alpha[p][m]
  float** _b;     // _b[p,q]=b(p,q)
  float* _lambda_i;     // _lambda_i[p]
  float* _lambda_d;     // _lambda_d[p]

  float* _fstat;

  float* _r0; // init distrib
  float* _s0;

  double _poutlier;
  double* _poutlierlogit;
  int _ncatlogit;
  double _outlier_logita, _outlier_logitb;
  gsl_vector *_outlier_logit;
  float* _pobs_outlier; //_pobs_outlier[0] P(X_t=NA| 0_t=1)  _pobs_outlier[1] P(X_t=x| 0_t=1) _pobs_outlier[2] P(X_t<xmin| 0_t=1) _pobs_outlier[3] P(X_t>xmax| 0_t=1)

  double _count_outlier;
  double** _count_logitoutlier;
  double _count_sum_sd;

  double* _count_sum_rhosd_0;
  double* _count_sum_rhosd_1;
  double* _count_sum_rhosd_2;
  double* _count_sum_rhosd_3;

  int _count_t;
  double** _count_alpha;
  double* _count_i12;
  double* _count_i1;
  double* _count_d12;
  double* _count_d1;

  double** _count_bpq;
  
  
 public:
  Params(int nS, int nR, float xmin, float xmax, float smin, float smax, float* alphamin, float maxdriftamplitude); 
  Params(ifstream& in); 
  ~Params();
  void initdeterministic(int type_sigma, int type_rho, int outlier_type);
  void initrandom(int type_sigma, int type_rho, int outlier_type, gsl_rng * r);
  void initrandom(gsl_rng * r);
  void definecatmu(double frac_cat_mu, Data** datav, int ndata);
  void definefstatkdens(Data** datav, int ndata, bool etaflat);
  void reset_counts();
  void force_alphamin(int p);
  void m_step(int iter);
  void print(ofstream& out);
  void alloclogit(int ncatlogit);
  void copycatmufstat(Params &templ);
  void compute_maxlambda();
  void force_maxlambda(int p);
};

double compute_loglik_poutlier(const gsl_vector *x, void *params);

#endif
